/*******************************************************************************
**  FILE:	    mesh.h
**  DESCRIPTION:    Base mesh class for simulation framework.  This class handles
**		    all of the memory management and OpenCL API calls to transfer
**		    and handle the manipulate the data on the GPU.
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#ifndef MESH_H
#define MESH_H

#include <QObject>
#include <QtGui>
#include "QtOclMesh_global.h"
#include "point3d.h"
#include "size3d.h"

// FORWARD DECLARATIONS
class MeshItem;
class NodeDescription;
class MeshEngine;


/**********************************************************
**
**  MESH
**
**********************************************************/

class QTOCLMESHSHARED_EXPORT Mesh : public QObject
{
  Q_OBJECT

protected:
  MeshEngine*         meshEngine_;
  bool                meshIs3D_;
  QList<MeshItem*>    supportedMeshItems_;
  QList<MeshEngine*>  supportedMeshEngines_;

private:
  QReadWriteLock      meshMutex_;
  QList<MeshItem*>    meshItems_;
  QList<MeshItem*>    selectedMeshItems_;
  Size3D<float>       meshResolution_;


public:
  explicit Mesh(QObject *parent = 0);
  ~Mesh();

  // GETTERS

  void setMesh3D(const bool meshIs3D);
  const bool isMesh3D() const;
  QReadWriteLock& getMeshMutex();
  void setMeshResolution(const Size3D<float> resolution);
  const Size3D<float> getMeshResolution() const;

  QList<MeshItem*> items() const;

  template <class T> const QList<T*>
  filteredItems() const
  {
    QList<T*> filteredList;
    T* filter;
    foreach(MeshItem* item, items()){
      filter = qobject_cast<T*>(item);
      if(filter != 0){
        filteredList.append(filter);
      }
    }
    return filteredList;
  }

  // SETTERS

  // MODIFIERS

  void addSupportedItem(MeshItem* meshItem);
  void removeSupportedItem(MeshItem* meshItem);
  QList<MeshItem*> getSupportedItems() const;

  void addSupportedEngine(MeshEngine* meshEngine);
  void removeSupportedEngine(MeshEngine* meshEngine);
  QList<MeshEngine*> getSupportedEngines() const;


  void selectItem( MeshItem* meshItem );
  void deselectItem( MeshItem* meshItem );
  void clearSelection();
  QList<MeshItem*> selectedItems() const;


private:

  void addItem( MeshItem* meshItem );
  void removeItem( MeshItem* meshItem );


signals:

  void modified(const Mesh* mesh);
  void updateEditor(Mesh* mesh);
  void meshItemsModified(Mesh* mesh);
  void selectionChanged(QList<MeshItem*> items);
  void itemInserted(const MeshItem* meshItem);
  void meshEngineChanged( MeshEngine* meshEngine);

public slots:

  void setItemsPos( const Point3D<float> pos);
  void adjustItemsPos( const Point3D<float> deltaPos);
  void setItemsSize( Size3D<float> size);
  void adjustItemsSize( Size3D<float> deltaSize);
  MeshItem* insertItem( const MeshItem* meshItem);
  void deleteItems( const QList<MeshItem*> meshItems);
};

QTOCLMESHSHARED_EXPORT QDataStream&
operator<<(QDataStream &out, const Mesh* mesh);

QTOCLMESHSHARED_EXPORT QDataStream&
operator>>(QDataStream &in, Mesh* mesh);


#endif // MESH_H
