/*******************************************************************************
**  FILE:	    mesh.cpp
**  DESCRIPTION:    Base mesh class for simulation framework.  This class handles
**		    all of the memory management and OpenCL API calls to transfer
**		    and handle the manipulate the data on the GPU.
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include <QObject>

#include "../inc/mesh.h"
#include "../inc/meshitems.h"
#include "../inc/point3d.h"
#include "../inc/meshengine.h"



//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


Mesh::Mesh(QObject *parent)
  : QObject     ( parent )
  , meshIs3D_ (false)
{
  // Mesh resolution of 1mm x 1mm x 1mm
  meshResolution_.setWidth(0.001);
  meshResolution_.setHeight(0.001);
  meshResolution_.setDepth(0.001);

}



Mesh::~Mesh()
{
  qDeleteAll(getSupportedItems());
  qDeleteAll(getSupportedEngines());
}



//-----------------------------------------------------------------------------
// OPERATOR OVERLOADS
//-----------------------------------------------------------------------------





//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------


void Mesh::setMesh3D(const bool meshIs3D)
{
  meshIs3D_ = meshIs3D;
}


const bool
Mesh::isMesh3D() const
{
  return meshIs3D_;
}



QReadWriteLock&
Mesh::getMeshMutex()
{
  return meshMutex_;
}



void
Mesh::setMeshResolution(const Size3D<float> resolution)
{
  meshResolution_ = resolution;
}



const Size3D<float>
Mesh::getMeshResolution() const
{
  return meshResolution_;
}



QList<MeshItem*>
Mesh::items() const
{
  return meshItems_;
}



// Modifiers

void
Mesh::addItem( MeshItem *meshItem )
{
  if( meshItems_.contains(meshItem) == false ){
    meshItem->setMesh(this);
    meshItems_.append(meshItem);
    emit modified(this);
    emit meshItemsModified(this);
    emit updateEditor(this);
  }
  else{
    qDebug()<<"Cannot add item, Item has already been added to mesh";
  }
}



void
Mesh::removeItem(MeshItem *meshItem)
{
  if( meshItems_.contains(meshItem) == true ){
    deselectItem(meshItem);
    MeshItem* item = meshItems_.takeAt(meshItems_.indexOf(meshItem));
    delete item;
    emit modified(this);
    emit meshItemsModified(this);
    emit updateEditor(this);
  }
  else{
    qDebug()<<"Cannot remove item, item has not been added to mesh";
  }
}



void
Mesh::selectItem(MeshItem *meshItem)
{
  if( selectedMeshItems_.contains(meshItem) == false ){
    selectedMeshItems_.append(meshItem);
    meshItem->setSelected(true);
    emit updateEditor(this);
    emit selectionChanged(selectedMeshItems_);
  }
  else{
    qDebug()<<"Item is already selected.";
  }
}



void
Mesh::deselectItem(MeshItem *meshItem)
{
  if( selectedMeshItems_.contains(meshItem) == true ){
    meshItem->setSelected(false);
    selectedMeshItems_.removeOne(meshItem);
    emit updateEditor(this);
    emit selectionChanged(selectedMeshItems_);
  }
  else{
    qDebug()<<"Cannot deselect item that is not selected.";
  }
}



void
Mesh::clearSelection()
{
  foreach(MeshItem* meshItem, selectedMeshItems_){
    meshItem->setSelected(false);
  }
  selectedMeshItems_.clear();
  emit updateEditor(this);
  emit selectionChanged(selectedMeshItems_);
}



QList<MeshItem*> Mesh::selectedItems() const
{
  return selectedMeshItems_;
}




void
Mesh::addSupportedItem(MeshItem *meshItem)
{
  if(!(supportedMeshItems_.contains(meshItem))){
   supportedMeshItems_.append(meshItem);
  }
}



void
Mesh::removeSupportedItem(MeshItem *meshItem)
{
  if(supportedMeshItems_.contains(meshItem)){
   supportedMeshItems_.removeOne(meshItem);
  }
}



QList<MeshItem*>
Mesh::getSupportedItems() const
{
//  QList<MeshItem*> tmpList;
//  tmpList.append(supportedBoundaryTypes_);
//  tmpList.append(supportedMaterialTypes_);
//  tmpList.append(supportedProbeTypes_);
//  tmpList.append(supportedSourceTypes_);
//  return tmpList;
  return supportedMeshItems_;
}




void
Mesh::addSupportedEngine(MeshEngine *meshEngine)
{
 if(!(supportedMeshEngines_.contains(meshEngine))){
  supportedMeshEngines_.append(meshEngine);
 }
}



void
Mesh::removeSupportedEngine(MeshEngine *meshEngine)
{
  if(supportedMeshEngines_.contains(meshEngine)){
   supportedMeshEngines_.removeOne(meshEngine);
  }
}



QList<MeshEngine*> Mesh::getSupportedEngines() const
{
  return supportedMeshEngines_;
}




//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
Mesh::setItemsPos(const Point3D<float> pos)
{
  Point3D<float> newPos = pos;
  if(!isMesh3D()){
    newPos.rz() = 0;
  }
  foreach(MeshItem* item, selectedItems()){
    item->setPos(newPos);
  }
  emit meshItemsModified(this);
  emit updateEditor(this);
}



void
Mesh::adjustItemsPos(const Point3D<float> deltaPos)
{
  Point3D<float> newDeltaPos = deltaPos;
  if(!isMesh3D()){
    newDeltaPos.rz() = 0;
  }
  foreach(MeshItem* item, selectedItems()){
    item->adjustPos(newDeltaPos);
  }
  emit meshItemsModified(this);
  emit updateEditor(this);
}



void
Mesh::setItemsSize(Size3D<float> size)
{
  Size3D<float> newSize = size;
  foreach(MeshItem* item, selectedItems()){
    if(!isMesh3D()){
      newSize.setDepth(item->minimumSize().depth());
    }
    item->setSize(newSize);
  }
  emit meshItemsModified(this);
  emit updateEditor(this);
}



void
Mesh::adjustItemsSize(Size3D<float> deltaSize)
{
  Size3D<float> newDeltaSize = deltaSize;
  if(!isMesh3D()){
    newDeltaSize.setDepth(0);
  }
  foreach(MeshItem* item, selectedItems()){
    item->adjustSize(newDeltaSize);
  }
  emit meshItemsModified(this);
  emit updateEditor(this);
}



MeshItem* Mesh::insertItem(const MeshItem *meshItem)
{
  clearSelection();
  MeshItem* newMeshItem = meshItem->clone();

  if(!isMesh3D()){
    newMeshItem->setUp(Point3D<float>(0,0,1));
    Point3D<float> itemPos = newMeshItem->pos();
    itemPos.rz() = 0;
    newMeshItem->setPos(itemPos);
  }

  blockSignals(true);
  addItem(newMeshItem);
  selectItem(newMeshItem);
  blockSignals(false);
  emit modified(this);
  emit meshItemsModified(this);
  emit updateEditor(this);

  return newMeshItem;
}



void Mesh::deleteItems(const QList<MeshItem *> meshItems)
{
  if(meshItems.count() > 0 ){
    foreach(MeshItem* item, meshItems){
      removeItem(item);
    }
    emit modified(this);
    emit meshItemsModified(this);
    emit updateEditor(this);
  }
}





//-----------------------------------------------------------------------------
// STREAM OPERATORS
//-----------------------------------------------------------------------------




QDataStream&
operator<<(QDataStream &out, const Mesh * mesh){
  QList<MeshItem*> meshItems = mesh->items();
  out << mesh->isMesh3D()
      << mesh->getMeshResolution()
      << meshItems.count();

  QString className;
  for(int i = 0; i< meshItems.count(); ++i){
    className = meshItems.at(i)->metaObject()->className();
    out << className << *(meshItems.at(i));
  }

  return out;
}



QDataStream&
operator>>(QDataStream &in, Mesh * mesh){
  Size3D<float> tmpRes;
  bool tmpIsMesh3D;
  int itemCount;

  in  >> tmpIsMesh3D
      >> tmpRes
      >> itemCount;

  // Set mesh Settings
  mesh->setMesh3D(tmpIsMesh3D);
  mesh->setMeshResolution(tmpRes);

  // Insert saved items

  QList<MeshItem*> supportedItems = mesh->getSupportedItems();

  QString className;
  MeshItem* tmpItem;
  mesh->blockSignals(true);
  for(int i = 0; i < itemCount; ++i){
    in >> className;

    foreach(MeshItem* baseItem, supportedItems){
      if(baseItem->metaObject()->className() == className){
        tmpItem = baseItem->clone();
        break;
      }
    }
    tmpItem->setMesh(mesh);
    in >> *tmpItem;

    // Inserting the item actually clones the item
    mesh->insertItem(tmpItem);
    // temporary item must be deleted
    delete tmpItem;
  }
  mesh->blockSignals(false);
  mesh->clearSelection();

  return in;
}






