/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "../inc/tlmengine2dgpu.h"
#include "../inc/oclmanager.h"


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------



TLMEngine2DGPU::TLMEngine2DGPU(Mesh* initMesh)
  : TLMEngine2D(initMesh)
  , probeBufferSize_ (100)
  , hostProbeBuffer_ (0)
{
    // Directory Navigation to the Kernel File
    QDir kernelDirPath = QDir(qApp->applicationDirPath());
    //kernelDirPath.cdUp();  // PPMS - Commented out by Poman
    kernelDirPath.cdUp();
    kernelDirPath.cd(QString("QtOclMesh"));
    kernelDirPath.cd(QString("kernels"));
    QString kernelPath =kernelDirPath.path().
                        append(QString("/dielectricmeshCL.cl"));
    // Create the OCL Manager with the specified Kernel File
    oclManager_ = new OCLManager(kernelPath);
//    oclManager_->setProfilingEnabled(true);
    setCompatibility(oclManager_->isHostCompatible());
    engineName_ = tr("2D TLM Shunt Node (GPU)");
    oclManager_->setProfilingEnabled(false);
}



TLMEngine2DGPU::TLMEngine2DGPU(const TLMEngine2DGPU &other):
    TLMEngine2D(other)
{
    oclManager_       = other.oclManager_;
    probeBufferSize_  = other.probeBufferSize_;
    setCompatibility(oclManager_->isHostCompatible());
}



TLMEngine2DGPU::~TLMEngine2DGPU()
{
  if(oclManager_!=0){
    delete oclManager_;
  }
  if(hostProbeBuffer_ != 0)
    delete[] hostProbeBuffer_;
}




//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------



TLMEngine2DGPU&
TLMEngine2DGPU::operator =(const TLMEngine2DGPU& other)
{
  if(&other!=this){
    TLMEngine2D::operator =(other);
    oclManager_ = other.oclManager_;
    setCompatibility(oclManager_->isHostCompatible());
  }
  return *this;
}



MeshEngine*
TLMEngine2DGPU::clone() const
{
    return new TLMEngine2DGPU(*this);
}



void
TLMEngine2DGPU::createSimulation()
{
  MeshEngine::createSimulation();
  meshSize_.setDepth(1);
  if(isCompatible()){
    hostProbeBuffer_ = new float[meshProbes_.count()*probeBufferSize_];
    startDataRead();
    oclManager_->copyBoundariesToDevice(nodeBoundaries_,
                                        nodeBoundaryCount_ * meshSize_.area() );
    oclManager_->copyBoundaryTypeTableToDevice(boundaryTypeTable_,
                                               numBoundaryTypes_ );
    oclManager_->copyPropertiesToDevice(nodeProperties_,
                                        nodePropertyCount_ * meshSize_.area() );
    oclManager_->createVoltageBuffer(nodeVoltages_,
                                     nodeVoltageCount_ * meshSize_.area() );
    copyProbesToDevice();
    copySourcesToDevice();

    endDataRead();
//    oclManager_->setupKernels( meshSize_, nodeVoltageCount_ * sizeof(float) );
    oclManager_->setupScatteringKernel(meshSize_
                                       ,nodeVoltageCount_ * sizeof(float)
                                       ,MeshSize(16,16));
    oclManager_->setupImpulseInterchangeKernel(meshSize_
                                               ,(nodeVoltageCount_-1) * sizeof(float)
                                               ,MeshSize(8,8));
    oclManager_->setupExcitationKernel(meshSize_);
    oclManager_->setupSamplingKernel(meshSize_);

    lastProbeUpdateIteration_ = -1;
  }
}



void
TLMEngine2DGPU::simulateIteration(int iteration)
{
  if(isCompatible()){
    MeshEngine::simulateIteration(iteration);
    exciteMesh(iteration);
    scatterImpulses();
    impulseInterchange();
    sampleMesh(iteration);
  }
}



void
TLMEngine2DGPU::updateHost()
{
  int iteration = getCurrentIteration();
  oclManager_->oclFinish();
  startDataWrite();
  oclManager_->copyVoltagesToHost(nodeVoltages_, meshSize_.area()*nodeVoltageCount_);
  updateHostProbes(iteration);
  endDataWrite();
  MeshEngine::updateHost();
  qDebug()<<"Excitation Time:"
          <<(float)oclManager_->totalExcitationKernelRunTime()/iteration
          <<" ns";
  qDebug()<<"Sampling Time:"
          <<(float)oclManager_->totalSamplingKernelRunTime()/iteration
          <<" ns";
  qDebug()<<"Scattering Time:"
          <<(float)oclManager_->totalScatteringKernelRunTime()/iteration
          <<" ns";
  qDebug()<<"Interchange Time:"
          <<(float)oclManager_->totalInterchangeKernelRunTime()/iteration
          <<" ns";

}



//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------



void
TLMEngine2DGPU::allocateResources(const MeshSize &meshSize)
{
  TLMEngine2D::allocateResources(meshSize);
}



//void
//TLMEngine2DGPU::initializeResources(const MeshSize &meshSize)
//{
//  TLMEngine2D::initializeResources(meshSize);
//}



void
TLMEngine2DGPU::releaseResources()
{
  if(hostProbeBuffer_ != 0)
    delete[] hostProbeBuffer_;

  hostProbeBuffer_ = 0;
  oclManager_->releaseBuffers();
  TLMEngine2D::releaseResources();
}



void
TLMEngine2DGPU::scatterImpulses()
{
  oclManager_->scatterImpulses();
}



void
TLMEngine2DGPU::impulseInterchange()
{
  oclManager_->impulseInterchange();
}



void
TLMEngine2DGPU::exciteMesh(int iteration)
{
  if(meshSources_.count() > 0){
    oclManager_->exciteMesh( iteration );
  }
}



void
TLMEngine2DGPU::sampleMesh(int iteration)
{
  if(meshProbes_.count()>0){
    oclManager_->sampleMesh( iteration );
    if(iteration - lastProbeUpdateIteration_ == probeBufferSize_){
      oclManager_->oclFinish();
      updateHostProbes(iteration);
    }
  }
}


//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------


void
TLMEngine2DGPU::updateHostProbes(int iteration)
{
  if(meshProbes_.count()>0){
    oclManager_->copyProbesToHost(hostProbeBuffer_,meshProbes_.count()*probeBufferSize_);
    int sampleIndex;
    ProbeMeshItem* probe;
    for(int p = 0; p < meshProbes_.count(); ++p){
      probe = meshProbes_.at(p);
      for(int i = lastProbeUpdateIteration_ + 1; i <= iteration; ++i){
        sampleIndex = (p * probeBufferSize_) + i % probeBufferSize_;
        probe->appendData(hostProbeBuffer_[sampleIndex]);
      }
    }
    lastProbeUpdateIteration_ = iteration;
  }
}



void
TLMEngine2DGPU::copyProbesToDevice()
{

    QVector<unsigned int> probeInfoVector;
    int numProbes = meshProbes_.count();
    foreach(ProbeMeshItem* probe, meshProbes_){
      probeInfoVector << packProbeInformation(probe);
    }
    oclManager_->copyProbesToDevice(numProbes,probeInfoVector,probeBufferSize_);
}



void
TLMEngine2DGPU::copySourcesToDevice()
{
    QVector<unsigned int> sourceInfoVector;
    QVector<float> sourceDataVector;
    int numSources = meshSources_.count();
    if(numSources>0){
      size_t dataOffset = 0;
      QVector<unsigned int> tmpInfoVector;
      foreach(SourceMeshItem* source,meshSources_){
        tmpInfoVector = packSourceInformation(source);
        tmpInfoVector<< dataOffset;
        dataOffset += tmpInfoVector.at(6); // Get the duration of the current source
        sourceInfoVector<<tmpInfoVector;
        sourceDataVector<<source->getData();
      }
    }
    oclManager_->copySourcesToDevice(numSources,sourceInfoVector,sourceDataVector);
}



QVector<unsigned int>
TLMEngine2DGPU::packSourceInformation(SourceMeshItem *source) const
{
  QVector<unsigned int> infoVector;
  // Source Position
  MeshPoint meshPoint = discretizePoint(source->pos());
  infoVector << meshPoint.cx();
  infoVector << meshPoint.cy();
  infoVector << meshPoint.cz();
  // Source size
  infoVector << 0; //source->size().width();
  infoVector << 0; //source->size().height();
  infoVector << 0; //source->size().depth();
  // Source duration
  if(source->getPropertyValue("SourceMeshItem/duration").toUInt()==0){
    infoVector << 1;
  }
  else{
    infoVector << source->getPropertyValue("SourceMeshItem/duration").toUInt();
  }
  infoVector << 0; //getDirectionCode(); // Todo: direction specification
  return infoVector;
}



QVector<unsigned int>
TLMEngine2DGPU::packProbeInformation(ProbeMeshItem *probe) const
{
  QVector<unsigned int> infoVector;
  // Source Position
  MeshPoint meshPoint = discretizePoint(probe->pos());
  infoVector << meshPoint.cx();
  infoVector << meshPoint.cy();
  infoVector << meshPoint.cz();

  return infoVector;
}

