/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "dataviewer.h"
#include <QThread>
#include <qtconcurrentrun.h>
#include "datatransforms.h"
#include "mesh.h"

//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------



DataViewer::DataViewer(const Mesh* mesh,QWidget *parent)
  : QWidget(parent)
  , graphScene_ (new QGraphicsScene(this))
  , graphView_ (new QGraphicsView(graphScene_,this))
  , transform_ (0)
  , transforming_ (false)
  , zoomFactor_ (1.0)
  , mesh_ (mesh)
  , mainLayout_ (new QGridLayout(this))
{

  graphScene_->setSceneRect(0,0,1024,768);

  transforms_.append(new DataTransform(this));  // Default Time Domain
  transforms_.append(new DFTransform(this));    // Discrete Fourier Transform

  createContextMenu();

  createTransformCombo();

  viewerToolLayout_ = new QHBoxLayout;
  viewerToolLayout_->addWidget(transformCombo_);
  viewerToolLayout_->addWidget(transformSettingsButton_);

  mainLayout_->addLayout(viewerToolLayout_,0,0);
  mainLayout_->addWidget(graphView_,1,0,1,4);

  changeTransform(0); // Select first transform
}



DataViewer::~DataViewer()
{
  graphScene_->clear();
  delete graphView_;
  delete graphScene_;

}


//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------

void
DataViewer::setData()
{
  rootData_.setData();
  updateGraph();
}


void
DataViewer::setData(QVector<float>& data)
{
  rootData_.setData(data);
  updateGraph();
}



const Mesh*
DataViewer::mesh() const
{
  return mesh_;
}





//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------


void
DataViewer::createTransformCombo()
{
  transformCombo_ = new QComboBox(this);
  QVariant data;
  foreach(DataTransform* transform, transforms_){
    data.setValue((void*)transform);
    transformCombo_->addItem(transform->getDisplayText(),data);
  }
  connect(transformCombo_, SIGNAL(currentIndexChanged(int)),
          this, SLOT(changeTransform(int)));

  transformSettingsButton_ = new QPushButton(tr("Settings..."));
  connect(transformSettingsButton_,SIGNAL(pressed()),
          this,SLOT(changeTransformSettings()));
}



void
DataViewer::resizeEvent(QResizeEvent *event)
{
  resizeView();
}



void
DataViewer::changeEvent(QEvent *event)
{
  resizeView();
}



void
DataViewer::contextMenuEvent(QContextMenuEvent *event)
{
  if(transformedData_.data().count() > 0){
    contextMenu_->exec(event->globalPos());
  }
}



void
DataViewer::graph(QGraphicsScene *scene)
{
  graphScene_->clear();
  QSizeF viewSize = scene->sceneRect().size();
  float normalizeY = viewSize.height()/2.0;
  float normalizeX = viewSize.width()/transformedData_.data().count();

  QPointF point,lastPoint;
  float rangeMax = transformedData_.range().y();
  for(int i = 0 ;i < transformedData_.data().count(); ++i){
    lastPoint = point;
    point.rx() = i*normalizeX;
    point.ry() = normalizeY*(1-transformedData_.data().at(i)/rangeMax);
    if(i>0)
      scene->addLine(lastPoint.x(),lastPoint.y(),point.x(),point.y());
  }
}



void
DataViewer::drawBackground(QGraphicsScene *scene)
{
  QFont font;
  font.setFamily("Sans Serif");
  font.setPointSize(10);

  QPen linePen(Qt::lightGray);
  QRectF sceneRect = scene->sceneRect();


  // Divisions within graph
  int yDivisions = 1;
  int xDivisions = 3;

  // Graph Padding for labels
  QPointF scenePixel = graphView_->mapToScene(1,1) - graphView_->mapToScene(0,0);

  float yPadding = 40;
  float xPadding = 40;


  // Insert Graph Title
  QGraphicsItem * text = scene->addText(transformedData_.title(),font);
  text->setPos((sceneRect.width()-text->boundingRect().width())/2.0,-yPadding);

  // Insert Y Axis Label
  text = scene->addText(transformedData_.yAxisTitle(),font);
  text->setPos(-xPadding,
               (sceneRect.height()+text->boundingRect().width())/2.0);
  text->rotate(-90);

  float yIncrement = sceneRect.height()/(yDivisions+1);
  float yDeltaAxisValue = (transformedData_.range().x()-transformedData_.range().y())/(yDivisions+1);
  float yAxisValue = transformedData_.range().y();
  for(int i = 0; i <= yDivisions+1; ++i){
    scene->addLine(0,i*yIncrement,sceneRect.width(),i*yIncrement,linePen);
    text = scene->addText(tr("%1").arg(yAxisValue+i*yDeltaAxisValue),font);
    text->rotate(-90);
    text->setPos(-text->boundingRect().height()/1.2,i*yIncrement+text->boundingRect().width()/2.0);

  }

  // Insert X Axis Label
  text = scene->addText(transformedData_.xAxisTitle(),font);
  text->setPos((sceneRect.width()-text->boundingRect().width())/2.0,
               sceneRect.height()+text->boundingRect().height()/1.5);

  float xIncrement = sceneRect.width()/(xDivisions+1);
  float xDeltaAxisValue = (transformedData_.domain().y()-transformedData_.domain().x())/(xDivisions+1);
  float xAxisValue = transformedData_.domain().x();
  for(int i = 0; i <= xDivisions+1; ++i){
    scene->addLine(i*xIncrement,0,i*xIncrement,sceneRect.height(),linePen);
    text = scene->addText(tr("%1").arg(xAxisValue+i*xDeltaAxisValue),font);
    text->setPos(i*xIncrement-text->boundingRect().width()/2.0,sceneRect.height());
  }


  graphPadding_.setX(xPadding*2);
  graphPadding_.setY(yPadding*2);

}





//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------

void
DataViewer::resizeView()
{
  QRectF sceneRect = graphScene_->sceneRect();
  float padX = graphPadding_.x();
  float padY = graphPadding_.y();
  sceneRect.setWidth(sceneRect.width()/zoomFactor_);
  sceneRect.setLeft(sceneRect.left()-padX);
  sceneRect.setTop(sceneRect.top()-padY/2.0);
  sceneRect.setBottom(sceneRect.bottom()+padY/2.0);

  graphView_->fitInView(sceneRect);

}


void
DataViewer::createContextMenu()
{
  contextMenu_ = new QMenu(tr("Options"),this);

  QAction* newAction = new QAction(tr("Save Data to file..."),contextMenu_);
  connect(newAction,SIGNAL(triggered()),this,SLOT(saveData()));
  contextMenu_->addAction(newAction);

  newAction = new QAction(tr("Save image..."),contextMenu_);
  connect(newAction,SIGNAL(triggered()),this,SLOT(saveImage()));
  contextMenu_->addAction(newAction);

}



void
DataViewer::writeFileHeader(QTextStream &out)
{
  out<<transform_->getDisplayText();
  out<<"\t"<<transform_->getToolTip();
  endl(out);
  endl(out);
  out.setFieldWidth(12);
  out.setRealNumberPrecision(10);
  out<<"NUM_POINTS\t"<<transformedData_.data().count();
  endl(out);
  out<<"RANGE\t"<<transformedData_.range().x()<<"\t"<<transformedData_.range().y();
  endl(out);
  out<<"DOMAIN\t"<<transformedData_.domain().x()<<"\t"<<transformedData_.domain().y();
  endl(out);
  out<<"DELTA_X\t"<<(transformedData_.domain().y()-transformedData_.domain().x())
        /transformedData_.data().count();
  endl(out);
  endl(out);
}


//-----------------------------------------------------------------------------
// PRIVATE SLOTS
//-----------------------------------------------------------------------------


void
DataViewer::draw()
{
  graph(graphScene_);
  drawBackground(graphScene_);
  resizeView();
  transforming_ = false;
}



void
DataViewer::changeTransform(int transformIndex)
{
  if(transform_ != 0){
    disconnect(transform_,SIGNAL(dataTransformed()),this,SLOT(draw()));
  }
  transform_ = (DataTransform*)(transformCombo_->itemData(transformIndex).value<void*>());
  connect(transform_,SIGNAL(dataTransformed()),this,SLOT(draw()));
  transformSettingsButton_->setEnabled(transform_->hasOptions());

  transforming_ = false;
  transformedData_.setData();
  updateGraph();
}


void
DataViewer::changeTransformSettings()
{
  DataTransformDialog* transformDialog = new DataTransformDialog(transform_);
  if(transformDialog->exec()){
    updateGraph();
  }
  delete transformDialog;
}


void
DataViewer::saveData()
{
  QFileDialog::Options options;
  QString selectedFilter = tr("Text Files (*.txt)");
  QString directory = QDesktopServices::storageLocation(QDesktopServices::DesktopLocation);
  QString fileName = QFileDialog::getSaveFileName(this,
                                                  tr("Save Data to file..."),
                                                  directory.append("\\untitled.txt"),
                                                  tr("All Files (*);;Text Files (*.txt)"),
                                                  &selectedFilter,
                                                  options);
  if (!fileName.isEmpty()){
    QFile dataFile(fileName);
    if(!dataFile.open(QIODevice::WriteOnly)){
      // ToDo: show dialog box informing user
      qDebug()<<"Could not open File.";
    }else{
      QTextStream out(&dataFile);
      writeFileHeader(out);
      int dataCount = transformedData_.data().count();
      out.setFieldWidth(12);
      out.setRealNumberPrecision(10);
      for(int i = 0; i < dataCount; ++i){
        out<<transformedData_.data().at(i);
        endl(out);
      }
      out.flush();
      dataFile.close();
    }
  }
}


void
DataViewer::saveImage()
{
  QFileDialog::Options options;
  QString selectedFilter = tr("Images (*.png *.bmp *.jpg)");
  QString directory = QDesktopServices::storageLocation(QDesktopServices::DesktopLocation);
  QString fileName = QFileDialog::getSaveFileName(this,
                                                  tr("Save Image to file..."),
                                                  directory.append("\\untitledImage.png"),
                                                  tr("Images (*.png *.jpg)"),
                                                  &selectedFilter,
                                                  options);
  if (!fileName.isEmpty()){
    QImage img(graphView_->width(),graphView_->height(),
               QImage::Format_ARGB32_Premultiplied);
    QPainter painter(&img);
    graphView_->render(&painter);
    painter.end();
    img.save(fileName);
  }

}

//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
DataViewer::updateGraph()
{
  if(!transforming_ && transform_ != 0){
    transforming_ = true;
    transform_->transformData(rootData_, transformedData_);
  }

}


