/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "gldisplaywidget.h"
#include <QtGui>


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


GLDisplayWidget::GLDisplayWidget(QWidget *parent)
    : QGLWidget (QGLFormat(QGL::DoubleBuffer | QGL::DepthBuffer), parent)
    , blockGLUpdate_ (false)
{
  setAutoBufferSwap(false);
}



GLDisplayWidget::~GLDisplayWidget()
{
  qDeleteAll(lights_);

}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------



void
GLDisplayWidget::setViewPortSize(const QSize newSize)
{
  viewPortSize_ = newSize;
}


const QSize
GLDisplayWidget::viewPortSize() const
{
  return viewPortSize_;
}




void GLDisplayWidget::initialize()
{
  initializeGL();
}



void
GLDisplayWidget::initGL()
{
  glClearColor(1.0, 1.0, 1.0, 0.0);

  // ALIASING AND FILLING SETTINGS
  glShadeModel(GL_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_POINT_SMOOTH);
  glFrontFace(GL_CCW);
}



void
GLDisplayWidget::setupGLLightingModel()
{
  // NO LIGHTING MODEL BY DEFAULT
  if(lights_.count() > 0){
    qWarning()<<"Warning: Lights have been defined without a lighting model,"
        "This is most likely due to a derrived ThreadedGLWidget which has not"
        "overloaded the setupGLLightingModel()";
  }
}



QList<GLLight*>
GLDisplayWidget::getLights() const
{
  return lights_;
}



//-----------------------------------------------------------------------------
// PROTECTED METHODS
//-----------------------------------------------------------------------------



void
GLDisplayWidget::initializeGL()
{
    if(lights_.count()>0){
      glEnable(GL_LIGHTING);
      foreach(GLLight* light, lights_){
        glEnable(light->glLight);
        glLightfv(light->glLight, GL_DIFFUSE, light->diffuse);
        glLightfv(light->glLight, GL_SPECULAR, light->specular);
        glLightfv(light->glLight, GL_AMBIENT, light->ambient);
      }
      //glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,0);
    }
    setupGLLightingModel();
    initGL();
}



void
GLDisplayWidget::closeEvent(QCloseEvent *event)
{
  // Close the widget
  QGLWidget::closeEvent(event);
}



void
GLDisplayWidget::resizeGL(int w, int h)
{
  viewPortSize_ = QSize(w,h);
  resizeViewPort(viewPortSize_);
}



void
GLDisplayWidget::paintGL()
{
  if(blockGLUpdate_ == false){
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    draw();
    glFinish();
    swapBuffers();
  }
}



void
GLDisplayWidget::addLight(GLLight *light)
{
  if(!lights_.contains(light)){
    lights_.append(light);
  }
}



void
GLDisplayWidget::removeLight(GLLight *light)
{
  if(lights_.contains(light)){
    lights_.removeOne(light);
  }
}



void GLDisplayWidget::resizeViewPort(QSize size)
{
  // SET UP THE PROJECTION MATRIX
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  setupViewingVolume(size);

  glMatrixMode(GL_MODELVIEW);
  glViewport(0, 0, size.width(), size.height());

  emit viewPortResized(size);
  updateGL();
}



GLint
GLDisplayWidget::pick(const QPoint& pos) const
{
  const int HIT_BUFFER_SIZE = 512;
  GLuint hitBuffer[HIT_BUFFER_SIZE];
  GLint viewport[4];

  glGetIntegerv(GL_VIEWPORT, viewport);
  glSelectBuffer(HIT_BUFFER_SIZE, hitBuffer);
  // enter select mode
  glRenderMode(GL_SELECT);

  glInitNames();
  glPushName(-1);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix((GLdouble)pos.x(),
                (GLdouble)(viewport[3] - pos.y()),
                5.0, 5.0, viewport);
  setupViewingVolume(viewPortSize_);
  draw();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();

  GLint hits = glRenderMode(GL_RENDER);

  if (hits == 0)
  {
      return -1;
  }

  int hitEntry;
  int closestName = hitBuffer[3];
  int closestDepth    = hitBuffer[1];
  for(int i = 1; i < hits; ++i){
    hitEntry = i*4;
    if(hitBuffer[hitEntry+1]<closestDepth){
      closestDepth = hitBuffer[hitEntry+1];
      closestName = hitBuffer[hitEntry+3];
    }
  }

  // return the name of the clicked surface
  return closestName;
}



//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------


void
GLDisplayWidget::blockGLUpdate(const bool block)
{
  blockGLUpdate_ = block;
}


