/*******************************************************************************
**  FILE:
**  DESCRIPTION:
**
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "openglscene.h"
#include <QtGui>
#include <QtOpenGL>
#ifndef GL_MULTISAMPLE
#define GL_MULTISAMPLE  0x809D
#endif

OpenGLScene::OpenGLScene(QMutex& initBufferMutex,QObject *parent) :
    QGraphicsScene(parent),
    modelDistance(0),
    modelRotation(Point3D<float>(0,0,0)),
    modelColor(QColor(Qt::red)),
    backgroundColor(QColor(Qt::black)),
    bufferMutex(&initBufferMutex),
    meshPoints(NULL)
{
    initialize();
}

OpenGLScene::~OpenGLScene(){
    //glDisable(GL_DEPTH_TEST);
}


void OpenGLScene::initialize(){
    glEnable(GL_DEPTH_TEST);
}

void OpenGLScene::drawBackground(QPainter *painter, const QRectF &rect){
    if (painter->paintEngine()->type() != QPaintEngine::OpenGL && painter->paintEngine()->type() != QPaintEngine::OpenGL2 ) {
        qDebug()<<painter->paintEngine()->type();
        qWarning("OpenGLScene: drawBackground needs a QGLWidget to be set as viewport on the graphics view");
        return;
    }

    // Clear the OpenGL screen to the default background color
    glClearColor(backgroundColor.redF(), backgroundColor.greenF(), backgroundColor.blueF(), 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);


    // Should do any transformations, lighting in here




    // Save current matrix
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();

    // Load the identity matrix
    glLoadIdentity();


    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    //glColor4f(modelColor.redF(), modelColor.greenF(), modelColor.blueF(), 1.0f);

    gluPerspective(70, width() / height(), 0.001, 1000);

    glTranslatef(0,0,-modelDistance);
    glRotatef(modelRotation.cx(), 1, 0, 0);
    glRotatef(modelRotation.cy(), 0, 1, 0);
    glRotatef(modelRotation.cz(), 0, 0, 1);

    glEnable(GL_MULTISAMPLE);
    renderMesh();
    glDisable(GL_MULTISAMPLE);

    glPopMatrix();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    //QTimer::singleShot(20, this, SLOT(update()));

}

void OpenGLScene::renderMesh(){
    if(meshPoints != NULL && meshIndices!=NULL){
	bufferMutex->lock();
	if(meshPoints->size()!=0 && meshIndices->size()!=0){
	    glEnableClientState(GL_VERTEX_ARRAY);
	    glEnableClientState(GL_COLOR_ARRAY);
	    glColorPointer(3, GL_FLOAT, 0, (float*)meshPointColors->data());
	    glVertexPointer(3, GL_FLOAT, 0, (float *)meshPoints->data());
	    glDrawElements(GL_LINE_STRIP,meshIndices->size(), GL_UNSIGNED_INT, meshIndices->data());
	    glDisableClientState(GL_COLOR_ARRAY);
	    glDisableClientState(GL_VERTEX_ARRAY);
	}
	bufferMutex->unlock();
    }
}

void OpenGLScene::updateBuffers(QVector<float> *points, QVector<int> *indices, QVector<float> *colors,MeshSize meshSize)
{
    if(meshPoints==NULL){
        modelDistance = sqrt(pow(meshSize.width(),2)/2+pow(meshSize.height(),2)/2);
        modelRotation.rx() = -25;
    }

    meshPoints = points;
    meshIndices = indices;
    meshPointColors = colors;
    update();
}

void OpenGLScene::setModelColor()
{
    const QColor color = QColorDialog::getColor(modelColor);
    if (color.isValid()) {
        modelColor = color;
        update();
    }
}

void OpenGLScene::setBackgroundColor()
{
    const QColor color = QColorDialog::getColor(backgroundColor);
    if (color.isValid()) {
        backgroundColor = color;
        update();
    }
}


void OpenGLScene::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    QGraphicsScene::mouseMoveEvent(event);
    if (event->isAccepted())
        return;
    const QPointF delta = event->scenePos() - event->lastScenePos();
    if (event->buttons() & Qt::LeftButton) {
        const Point3D<float> angularImpulse = Point3D<float>(delta.y(), delta.x(), 0) * 0.1;

        modelRotation += angularImpulse;
        event->accept();
    }else if (event->buttons() & Qt::MiddleButton) {
        modelRotation += Point3D<float>(0,0,delta.x())*0.1;
	event->accept();
    }
    update();
}

void OpenGLScene::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    QGraphicsScene::mousePressEvent(event);
    if (event->isAccepted())
        return;

    event->accept();
}

void OpenGLScene::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    QGraphicsScene::mouseReleaseEvent(event);
    if (event->isAccepted())
        return;

    event->accept();
    update();
}

void OpenGLScene::wheelEvent(QGraphicsSceneWheelEvent *event)
{
    QGraphicsScene::wheelEvent(event);
    if (event->isAccepted())
        return;
    modelDistance *= qPow(1.2, -event->delta() / 120);
    event->accept();
    update();
}
