/*******************************************************************************
**  FILE:           editor.cpp
**  DESCRIPTION:    Editor widget used in the Electrosim application.
**                  Responsible for providing the editor interface for both 2D
**                  and 3D meshes.
**
**  AUTHOR:         Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#include "editor.h"
//#include "meshitemmanager.h"
#include "qtoclmesh.h"
//#include "editorview.h"
#include "editorview2d.h"
#include "editorview3d.h"
#include "pointmappergl.h"
#include "meshitemselector.h"
#include "meshitemlistselector.h"
#include "editorstatemachine.h"
#include <QtGui>


//-----------------------------------------------------------------------------
// CONSTRUCTORS / DESTRUCTORS
//-----------------------------------------------------------------------------


Editor::Editor(Mesh* initMesh, QWidget *parent) :
    QWidget       (parent),
    mesh_         (initMesh),
    pointMapper_  (new PointMapperGL(1.0,0.0001,1)),
    stateMachine_ (new EditorStateMachine(this)),
    mainLayout    (new QGridLayout)
{
    setAttribute(Qt::WA_DeleteOnClose);
    setWindowProperties();
    setFocusPolicy(Qt::StrongFocus);

    setupGLViews();
    setupMeshItemDock();

    createEditorMenu();
    //createMenuBar();

    connectSignals();

    pointMapper_->setSceneDepths(Point3D<float>(0,0,0));

    readSettings();

    setLayout(mainLayout);
}



Editor::~Editor()
{
  if(pointMapper_!=0){
    delete pointMapper_;
  }
  if(stateMachine_!=0){
    delete stateMachine_;
  }
}



//-----------------------------------------------------------------------------
// PUBLIC METHODS
//-----------------------------------------------------------------------------



const Editor::EditorMode Editor::getEditorMode() const
{
  return editorMode_;
}



void
Editor::setEditorMode(Editor::EditorMode mode)
{
  editorMode_ = mode;
  EditorMode actionMode;
  foreach(QAction* action, editorModeMenu->actions()){
    actionMode = (Editor::EditorMode)action->data().toInt();
    action->blockSignals(true);
    action->setChecked(actionMode == editorMode_);
    action->blockSignals(false);
  }
  bool is3D = editorMode_ == Editor::Edit3D;
  mesh_->setMesh3D(is3D);
  set3DEditorEnabled(is3D);
  emit editorModeChanged(editorMode_);
}



QDockWidget* Editor::getMeshItemDockWidget() const
{
  return meshItemDock;
}



PointMapperGL&
Editor::getPointMapper() const
{
  return *pointMapper_;
}


Mesh&
Editor::getMesh() const
{
  return *mesh_;
}



EditorStateMachine*
Editor::getStateMachine() const
{
  return stateMachine_;
}



QMenu*
Editor::getEditorMenu() const
{
  return editorMenu;
}


QMenu*
Editor::getItemActionMenu() const
{
  return stateMachine_->getItemActionMenu();
}



//-----------------------------------------------------------------------------
// PRIVATE METHODS
//-----------------------------------------------------------------------------



void
Editor::setupGLViews()
{
  GLfloat cameraDistance = pointMapper_->worldSize()/2;


  viewSplitterH       = new QSplitter(Qt::Horizontal, this);
  viewSplitterVLeft   = new QSplitter(Qt::Vertical, viewSplitterH);
  viewSplitterVRight  = new QSplitter(Qt::Vertical, viewSplitterH);

  connect(viewSplitterVLeft,SIGNAL(splitterMoved(int,int)),
          this,SLOT(splitterMoved(int,int)));
  connect(viewSplitterVRight,SIGNAL(splitterMoved(int,int)),
          this,SLOT(splitterMoved(int,int)));



  Point3D<float> vAxis,hAxis;
  // SET UP THE TOP VIEW OF THE EDITOR
  hAxis = Point3D<float>(1,0,0);
  vAxis = Point3D<float>(0,0,-1);
  editorViewMap.insert("top", new EditorView2D(this,hAxis,vAxis));
  // Set up the camera position for the view
  editorViewMap.value(QString("top"))->setCameraDistance(cameraDistance);
  viewSplitterVLeft->addWidget(editorViewMap.value(QString("top")));


  // SET UP THE FRONT VIEW OF THE EDITOR
  hAxis = Point3D<float>(1,0,0);
  vAxis = Point3D<float>(0,1,0);
  editorViewMap.insert("front",new EditorView2D(this,hAxis,vAxis));
  // Set up the camera position for the view
  editorViewMap.value(QString("front"))->setCameraDistance(cameraDistance);
  viewSplitterVLeft->addWidget(editorViewMap.value(QString("front")));


  // SET UP THE 3D VIEW OF THE EDITOR
  editorViewMap.insert("3d", new EditorView3D(this));
  // Set up the camera position for the view
  viewSplitterVRight->addWidget(editorViewMap.value(QString("3d")));


  // SET UP THE RIGHT VIEW OF THE EDITOR
  hAxis = Point3D<float>(0,0,-1);
  vAxis = Point3D<float>(0,1,0);
  editorViewMap.insert("right", new EditorView2D(this,hAxis,vAxis));
  // Set up the camera position for the view
  editorViewMap.value(QString("right"))->setCameraDistance(cameraDistance);
  viewSplitterVRight->addWidget(editorViewMap.value(QString("right")));

  mainLayout->addWidget(viewSplitterH,0,0);



}



void
Editor::setupMeshItemDock()
{
  meshItemDock  = new QDockWidget(QString("Mesh Items"));
  meshItemDock->setObjectName(QString("meshItemDockWidget"));
  meshItemDock->setMinimumSize(QSize(180, 316));
  meshItemDock->setMaximumSize(QSize(180, 524287));
  meshItemDock->setLayoutDirection(Qt::LeftToRight);

  meshItemSelector_ = new MeshItemListSelector(getSupportedMeshItems());
  meshItemDock->setWidget(meshItemSelector_);

}



QList<MeshItem*>
Editor::getSupportedMeshItems() const
{  
  return mesh_->getSupportedItems();
}



void
Editor::connectSignals()
{
  EditorView2D* view2D = 0;
  EditorView3D* view3D = 0;
  foreach(EditorView* view, editorViewMap){
    view2D = qobject_cast<EditorView2D*>(view);
    view3D = qobject_cast<EditorView3D*>(view);
    if(view2D != 0){
        setEnabled2DEditorView(view2D,true);
    }
    else if(view3D != 0){
      connect(view3D, SIGNAL(mousePressed(QPoint,
                                          Qt::MouseButtons,
                                          Qt::KeyboardModifiers)),
              stateMachine_, SLOT(mousePressed3D(QPoint,
                                                 Qt::MouseButtons,
                                                 Qt::KeyboardModifiers)));
    }
  }


  connect(mesh_,SIGNAL(selectionChanged(QList<MeshItem*>)),
          meshItemSelector_,SLOT(syncItemViewProperties(QList<MeshItem*>)));
  connect(meshItemSelector_,SIGNAL(itemSelected(MeshItem*)),stateMachine_,
          SLOT(setCurrentItem(MeshItem*)));
  connect(meshItemSelector_,SIGNAL(propertyChanged(QString,QVariant)),
          stateMachine_,SLOT(changeItemProperty(QString,QVariant)));



  connect(stateMachine_,SIGNAL(itemReadyToInsert(const MeshItem*)),
          mesh_,SLOT(insertItem(const MeshItem*)));
  connect(stateMachine_,SIGNAL(setCursorPos(QPoint)),this,SIGNAL(setCursorPos(QPoint)));
//  connect(stateMachine_,SIGNAL(scaleItems(Size3D<float>)),
//          mesh_,SLOT(setItemSize(Size3D<float>)));
//  connect(stateMachine_,SIGNAL(moveItems(Point3D<float>)),
//          mesh_,SLOT(setItemPos(Point3D<float>)));

  connect(mesh_,SIGNAL(itemInserted(const MeshItem*)),
          meshItemSelector_,SLOT(clearItemViewSelection()));
  connect(mesh_,SIGNAL(itemInserted(const MeshItem*)),
          stateMachine_,SLOT(itemInserted(const MeshItem*)));


  foreach(EditorView* view, editorViewMap){
    connect(mesh_,SIGNAL(updateEditor(Mesh*)),
            view,SIGNAL(renderMeshItems(Mesh*)));
  }
}



void
Editor::setEnabled2DEditorView(EditorView2D *view, const bool enabled)
{
  if(enabled){
    connect(view, SIGNAL(mouseMoved(QPoint,
                                      Qt::MouseButtons,
                                      Qt::KeyboardModifiers)),
            stateMachine_, SLOT(mouseMoved2D(QPoint,
                                             Qt::MouseButtons,
                                             Qt::KeyboardModifiers)));
    connect(view, SIGNAL(mouseReleased(QPoint,
                                         Qt::MouseButtons,
                                         Qt::KeyboardModifiers)),
            stateMachine_, SLOT(mouseReleased2D(QPoint,
                                                Qt::MouseButtons,
                                                Qt::KeyboardModifiers)));
    connect(view, SIGNAL(mousePressed(QPoint,
                                        Qt::MouseButtons,
                                        Qt::KeyboardModifiers)),
            stateMachine_, SLOT(mousePressed2D(QPoint,
                                               Qt::MouseButtons,
                                               Qt::KeyboardModifiers)));
  }
  else{
      disconnect(view, SIGNAL(mouseMoved(QPoint,
                                           Qt::MouseButtons,
                                           Qt::KeyboardModifiers)),
                 stateMachine_, SLOT(mouseMoved2D(QPoint,
                                                  Qt::MouseButtons,
                                                  Qt::KeyboardModifiers)));
      disconnect(view, SIGNAL(mouseReleased(QPoint,
                                              Qt::MouseButtons,
                                              Qt::KeyboardModifiers)),
                 stateMachine_, SLOT(mouseReleased2D(QPoint,
                                                     Qt::MouseButtons,
                                                     Qt::KeyboardModifiers)));
      disconnect(view, SIGNAL(mousePressed(QPoint,
                                             Qt::MouseButtons,
                                             Qt::KeyboardModifiers)),
                 stateMachine_, SLOT(mousePressed2D(QPoint,
                                                    Qt::MouseButtons,
                                                    Qt::KeyboardModifiers)));
    }
}


void
Editor::set3DEditorEnabled(const bool enabled)
{
  foreach(EditorView* view, editorViewMap){
    EditorView2D* view2D = qobject_cast<EditorView2D*>(view);
    if(view2D != 0){
      if(enabled){
        setEnabled2DEditorView(view2D,true);
      }
      else{
        if(view2D->viewMask() != Point3D<float>(1,1,0)){
          setEnabled2DEditorView(view2D,false);
        }
      }
    }
  }
}



void
Editor::readSettings()
{
    QSettings settings;
    nodeBoundariesVisible_ = settings.value("meshEditor/nodeBoundariesVisible",QVariant(true)).toBool();
    nodesVisible_ = settings.value("meshEditor/nodesVisible",QVariant(true)).toBool();
    nodeColor_ = settings.value("meshEditor/nodeColor",QVariant(Qt::black)).value<QColor>();
    nodeBoundaryColor_ = settings.value("meshEditor/nodeBoundaryColor",QVariant(Qt::darkGray)).value<QColor>();
    backgroundColor_ = settings.value("meshEditor/backgroundColor",QVariant(Qt::white)).value<QColor>();
    emit nodeBoundariesVisibleChanged(nodeBoundariesVisible_);
    emit nodesVisibleChanged(nodesVisible_);
    emit nodeColorChanged(nodeColor_);
    emit nodeBoundaryColorChanged(nodeBoundaryColor_);
    emit backgroundColorChanged(backgroundColor_);
}



void
Editor::writeSettings()
{
    QSettings settings;

    settings.setValue("meshEditor/nodeBoundariesVisible",nodeBoundariesVisible_);
    settings.setValue("meshEditor/nodesVisible",nodesVisible_);
    settings.setValue("meshEditor/nodeColor",nodeColor_);
    settings.setValue("meshEditor/nodeBoundaryColor",nodeBoundaryColor_);
    settings.setValue("meshEditor/backgroundColor",backgroundColor_);
}



void
Editor::closeEvent(QCloseEvent *event)
{
    writeSettings();
//    if (maybeSave()) {
//        event->accept();
//    } else {
//        event->ignore();
//    }
    event->accept();
}




void
Editor::setWindowProperties()
{
    setWindowTitle(tr("Mesh Editor"));
    setWindowState(Qt::WindowMaximized);

    Qt::WindowFlags flags = 0;
    flags |= Qt::SubWindow;
    flags ^= Qt::WindowCloseButtonHint;
    flags |= Qt::WindowMinMaxButtonsHint;

    setWindowFlags(flags);

}



void
Editor::createEditorMenu()
{
    editorMenu = new QMenu(this);
    editorMenu->setTitle(tr("E&ditor"));

//    QMenu* appearanceMenu = new QMenu(tr("&Appearance"),editorMenu);

//    actionEditorNodesVisible = new QAction(QString("Nodes Visislbe"),this);
//    actionEditorNodesVisible->setCheckable(true);
//    actionEditorNodesVisible->setChecked(nodesVisible_);
//    appearanceMenu->addAction(actionEditorNodesVisible);
//    connect(actionEditorNodesVisible,SIGNAL(toggled(bool)),this,SLOT(changeNodesVisible(bool)));
//    connect(this,SIGNAL(nodesVisibleChanged(bool)),actionEditorNodesVisible,SLOT(setChecked(bool)));

//    actionEditorNodeBoundariesVisible = new QAction(QString("Node Boundaries Visislbe"),this);
//    actionEditorNodeBoundariesVisible->setCheckable(true);
//    actionEditorNodeBoundariesVisible->setChecked(nodeBoundariesVisible_);
//    appearanceMenu->addAction(actionEditorNodeBoundariesVisible);
//    connect(actionEditorNodeBoundariesVisible,SIGNAL(toggled(bool)),this,SLOT(changeNodeBoundariesVisible(bool)));
//    connect(this,SIGNAL(nodeBoundariesVisibleChanged(bool)),actionEditorNodeBoundariesVisible,SLOT(setChecked(bool)));

//    actionEditorNodeColor = new QAction(createColorIcon(nodeColor_),QString("Node Color"),this);
//    appearanceMenu->addAction(actionEditorNodeColor);
//    connect(actionEditorNodeColor,SIGNAL(triggered()),this,SLOT(changeNodeColor()));

//    actionEditorNodeBoundaryColor = new QAction(createColorIcon(nodeBoundaryColor_),QString("Node Boundary Color"),this);
//    appearanceMenu->addAction(actionEditorNodeBoundaryColor);
//    connect(actionEditorNodeBoundaryColor,SIGNAL(triggered()),this,SLOT(changeNodeBoundaryColor()));

//    actionEditorBackgroundColor = new QAction(createColorIcon(backgroundColor_),QString("Background Color"),this);
//    appearanceMenu->addAction(actionEditorBackgroundColor);
//    connect(actionEditorBackgroundColor,SIGNAL(triggered()),this,SLOT(changeBackgroundColor()));

//    editorMenu->addMenu(appearanceMenu);

    editorModeMenu = new QMenu(tr("Editing &Mode"),editorMenu);

    QAction* newAction = new QAction(QString("2D Mode"),editorModeMenu);
    QVariant actionData = QVariant::fromValue((int)Editor::Edit2D);
    newAction->setData(actionData);
    newAction->setCheckable(true);
    connect(newAction,SIGNAL(triggered()),this,SLOT(changeEditorMode()));
    editorModeMenu->addAction(newAction);

    newAction = new QAction(QString("3D Mode"),editorModeMenu);
    actionData = QVariant::fromValue((int)Editor::Edit3D);
    newAction->setData(actionData);
    newAction->setCheckable(true);
    connect(newAction,SIGNAL(triggered()),this,SLOT(changeEditorMode()));
    editorModeMenu->addAction(newAction);

    editorMenu->addMenu(editorModeMenu);

}



void
Editor::createMenuBar()
{
    editorMenuBar = new QMenuBar(this);

//    QAction* meshItemContextMenuAction = editorMenuBar->addMenu(meshEditor->getMeshItemContextMenu());
//    meshItemContextMenuAction->setText(tr("Edit"));

    QAction* editorViewMenuAction = editorMenuBar->addMenu(editorMenu);
    editorViewMenuAction->setText(tr("View"));

    editorMenuBar->setSizePolicy(QSizePolicy::Expanding,QSizePolicy::Fixed);
    // CHANGE THE LINE BELOW TO BE MORE DYNAMIC
    editorMenuBar->setFixedHeight(30);
    mainLayout->setMenuBar(editorMenuBar);
}



const QIcon
Editor::createColorIcon(const QColor& color)
{
    QPixmap colorPixmap(50,50);
    colorPixmap.fill(color);
    return QIcon(colorPixmap);
}




//-----------------------------------------------------------------------------
// PUBLIC SLOTS
//-----------------------------------------------------------------------------



void
Editor::changeNodesVisible(const bool visibility)
{
    nodesVisible_ = visibility;
    emit nodesVisibleChanged(nodesVisible_);
}



void
Editor::changeNodeBoundariesVisible(const bool visibility)
{
    nodeBoundariesVisible_ = visibility;
    emit nodeBoundariesVisibleChanged(nodeBoundariesVisible_);

}



void
Editor::changeNodeColor()
{
  QColor newNodeColor = QColorDialog::getColor(nodeColor_,this);
  if(newNodeColor.isValid()){
    nodeColor_ = newNodeColor;
    actionEditorNodeColor->setIcon(createColorIcon(nodeColor_));
    emit nodeColorChanged(nodeColor_);
  }
}



void
Editor::changeNodeBoundaryColor()
{
  QColor newBoundaryColor = QColorDialog::getColor(nodeBoundaryColor_,this);
  if(newBoundaryColor.isValid()){
    nodeBoundaryColor_ = newBoundaryColor;
    actionEditorNodeBoundaryColor->setIcon(createColorIcon(nodeBoundaryColor_));
    emit nodeBoundaryColorChanged(nodeBoundaryColor_);
  }
}



void
Editor::changeBackgroundColor()
{
  QColor newBackgroundColor = QColorDialog::getColor(backgroundColor_,this);
  if(newBackgroundColor.isValid()){
    backgroundColor_ = newBackgroundColor;
    actionEditorBackgroundColor->setIcon(createColorIcon(backgroundColor_));
    emit backgroundColorChanged(backgroundColor_);
  }
}



/******************************************************************
**  PRIVATE SLOTS
******************************************************************/


void
Editor::changeEditorMode()
{
  QAction* senderAction = qobject_cast<QAction*>(QObject::sender());
  if(senderAction != 0){
    QVariant newMode = senderAction->data();
    setEditorMode((Editor::EditorMode)newMode.toInt());
  }else{
    return;
  }

}



void
Editor::splitterMoved(int pos, int index)
{
  QSplitter* splitterPtr = qobject_cast<QSplitter*>(QObject::sender());

  if(splitterPtr == viewSplitterVLeft){
    viewSplitterVRight->setSizes(viewSplitterVLeft->sizes());
  }
  else if(splitterPtr == viewSplitterVRight){
    viewSplitterVLeft->setSizes(viewSplitterVRight->sizes());
  }

}




