/*******************************************************************************
**  FILE:	    editor.h
**  DESCRIPTION:    Editor widget used in the Electrosim application.  Responsible
**		    for providing the editor interface for both 2D and 3D meshes.
**
**  AUTHOR:	    Colter McQuay
**  ORGANIZATION:   University Of Victoria
**  SUPERVISOR:	    Dr. Poman So
**
*******************************************************************************/

#ifndef EDITOR_H
#define EDITOR_H

#include <QWidget>
#include <QtGui>
#include <QGLWidget>
#include "point3d.h"
#include "meshitems.h"


class MeshEditor;
class Mesh;
class PointMapperGL;
class EditorView;
class EditorView2D;
class EditorView3D;
class MeshItemSelector;
class EditorStateMachine;


class Editor : public QWidget
{
    Q_OBJECT
public:
  enum EditorMode{Edit2D,Edit3D};
private:
  Mesh*                 mesh_;
  PointMapperGL*        pointMapper_;
  MeshEditor*           meshEditor;
  MeshItem*             newItem_;
  EditorStateMachine*   stateMachine_;
  EditorMode            editorMode_;

  QString               curFileName_;

  bool            nodesVisible_;
  bool            nodeBoundariesVisible_;
  QColor          nodeColor_;
  QColor          nodeBoundaryColor_;
  QColor          backgroundColor_;


//-----------------------------------------------------------------------------
// USER INTERFACE ELEMENTS
//-----------------------------------------------------------------------------

  QGridLayout * mainLayout;

  QMap<QString, EditorView*> editorViewMap;
  QMap<QString,QScrollBar *> scrollBarMap;
  QSplitter * viewSplitterH;
  QSplitter * viewSplitterVLeft;
  QSplitter * viewSplitterVRight;


  QMenuBar *  editorMenuBar;
  QMenu *     editorMenu;
  QMenu *     editorModeMenu;
  QAction *   actionEditorNodesVisible;
  QAction *   actionEditorNodeBoundariesVisible;
  QAction *   actionEditorNodeColor;
  QAction *   actionEditorNodeBoundaryColor;
  QAction *   actionEditorBackgroundColor;

  QDockWidget* meshItemDock;
  MeshItemSelector* meshItemSelector_;


public:
  explicit Editor(Mesh* initMesh, QWidget *parent=0);
  ~Editor();

  const EditorMode getEditorMode() const;
  void setEditorMode(Editor::EditorMode mode);

  QDockWidget* getMeshItemDockWidget() const;
  PointMapperGL& getPointMapper() const;
  Mesh& getMesh() const;
  EditorStateMachine* getStateMachine() const;
  QMenu* getEditorMenu() const;
  QMenu* getItemActionMenu() const;


protected:
  void closeEvent(QCloseEvent *event);

private:
  void connectSignals();
  void setEnabled2DEditorView(EditorView2D* view, const bool enabled);
  void set3DEditorEnabled(const bool enabled);
  void readSettings();
  void writeSettings();
  void setWindowProperties();

  const QIcon createColorIcon(const QColor& color);

  void setupGLViews();
  void createEditorMenu();
  void createMenuBar();
  void setupMeshItemDock();

  QList<MeshItem*> getSupportedMeshItems() const;

signals:
  void editorModeChanged(Editor::EditorMode mode);
  void nodesVisibleChanged(const bool&);
  void nodeBoundariesVisibleChanged(const bool&);
  void nodeColorChanged(const QColor&);
  void nodeBoundaryColorChanged(const QColor&);
  void backgroundColorChanged(const QColor&);
  void drawingResolutionChanged(const float newResolution);
  void setCursorPos(QPoint pos);

public slots:
  void changeNodesVisible(const bool visibility);
  void changeNodeBoundariesVisible(const bool visibility);
  void changeNodeColor();
  void changeNodeBoundaryColor();
  void changeBackgroundColor();

private slots:
  void changeEditorMode();
  void splitterMoved(int pos, int index);

};


#endif // EDITOR_H
